/* BusSched.c
   BusSched is a utility for Palm OS based PDAs.
   It is designed to display Dublin Bus bus schedules. 

   $Id: BusSched.c,v 1.7 2006/08/22 21:46:22 daymobrew Exp $

 * (C) 2003-2006 Damien Carbery <daymobrew@users.sourceforge.net>
*/

/* This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <PalmOS.h>
#include <PalmUtils.h>
#include <TblGlue.h>

#include "BusSchedRsrc.h"

/* ***
  TODO
//  * Save/restore preferences. Date NOT being stored because user is likely to
//    want to see today's times.
//  * Page Up key = Earlier, Page Down key = Later.
  * PgUp/PgDn: If Earlier/Later button not visible then ignore request.
  * Put time in main form title, maybe include offset
     - "BusSched - 21:35 (-0:21)"
  * Allow multi-page route info.
  * Allow searching of BusSched databases. Ignore searches with just numbers.
  * Multi-page info form. Currently limitied to one page.
  * Consider putting version number in the code and appending it to the About 
    form when it is opened. This way the resources do not have to be changed 
    for each update.

  BUGS
//  * FIXED 1.4: Can delete current database!!
//  * CNR: Launch app, Select Database form, Cancel. No longer showing current buses!!
  * When 'No Service' alert clicked, table outline may not be drawn (ok on my device
    but not in simulator).
   ***
*/

/**********************************************************************
 * Global defines for this module
 **********************************************************************/
/* BusSched app and db IDs. */
#define CREATORID          'DCBS'   /* Damien Carbery Bus Schedule. */
#define PDBTYPE            'DATA'
#define PREFAPPVERSION     1        /* Version of the preferences structure. */

#define FIRST              0    /* Display first buses of day. */
#define NEXT              -1    /* Whether to display next buses, first buses or 'Later' buses. */
#define DIRECTION_OUT   false   /* Bus direction defines. */
#define DIRECTION_BACK   true

#define MAX_ROUTES        40    /* Max number of routes in a database. Arbitrary limit. */
#define MAX_ROUTE_NUM_LEN  6    /* Max length of route number (incl letters). Arbitrary limit. */
#define TIME_SEP_CHAR    ','    /* Character separating times in timetable. */
#define MAX_DAY_LENGTH     9    /* Max length of a day label string. */
                                /* The on-screen space isn't big.    */
#define BLANK_CELL        -1    /* Mark a table cell as blank. */

/* Route record offsets within group of ROUTE_REC_SIZE. */
#define ROUTE_NUM_OFFSET      0
#define FROM_OFFSET           1
#define TO_OFFSET             2
#define MF_OUT_TIMES_OFFSET   3
#define MF_BACK_TIMES_OFFSET  4
#define SA_OUT_TIMES_OFFSET   5
#define SA_BACK_TIMES_OFFSET  6
#define SU_OUT_TIMES_OFFSET   7
#define SU_BACK_TIMES_OFFSET  8
#define ROUTE_INFO_OFFSET     9
#define ROUTE_REC_SIZE       10    /* Number of records that constitute a route. */

/* For the memory block that stores non-numeric data for each table cell. */
#define CELL_DATA_LENGTH   4
#define CELL_DATA_SPACE    (CELL_DATA_LENGTH + 1)


/* To store program preferences. */
typedef struct
{
  UInt8 nDay;
  UInt16 nRoute;
  Boolean bDirection;
  UInt16 nMinsFromTerminus;    /* To look at buses that left Now - nMinsFromTerminus ago. */
  Char DBName[ dmDBNameLength ];  // TODO: Is this needed: PdbID and CardNo are better.
  LocalID PdbID;
  UInt16 CardNo;
} sPrefsStruct;

/* To store Database info in the SelectTimetableForm table. */
typedef struct {
    Char DBName[ dmDBNameLength ];
    LocalID PdbID;
    UInt16 CardNo;
} dbRowType;


/***********************************************************************
 * Global variables for this module
 **********************************************************************/
sPrefsStruct sPrefs;    /* Preferences structure. */
MemHandle PdbArrayH = NULL;    /* Pointer to array of dbRowType form timetable selection form. */
UInt16 gnCurrentTime = 0;    /* Store time so can show next buses. */
UInt16 nNextBusOffset = NEXT;
UInt16 nLastViewOffset = NEXT;  /* Used to restore timetable when returning from other forms. */
static MemHandle pTableData = NULL;  /* Memory block that stores non-numeric data for each table cell. */
static Char szEndPoint[ 30 ];    // TODO: Can this be done without??


/***********************************************************************
 * Function prototypes
 **********************************************************************/
#if 0
void DebugShowIntegerData( Coord x1, Coord y1, Coord x2, Coord y2 );    /* Display integer data in alert for debugging. */
void DebugShowLongData( UInt32 lNum );
void DebugStringAndNumber( Char * szString, UInt16 nNumber );
#endif

Err RomVersionCompatible( const UInt32 requiredVersion, const UInt16 launchFlags );
void StartApplication( void );
void StopApplication( void );	/* Clean up before app exit. */
void EventLoop( void );
Boolean ApplicationHandleEvent( const EventPtr event );	/* Handle form load events. */
Boolean MainFormHandleEvent( EventPtr event );
Int16 GetOffset( const TablePtr pTable, const Int16 nRow, const Int16 nColumn );
void TableDrawItemCallback( TableType *pTable, Int16 row, Int16 column, RectangleType* bounds );
void LoadMainFormTable( FormPtr pForm );
void UpdateCurrentTime( void );
Boolean SelectTimetableFormHandleEvent( EventPtr event );
Boolean PreferencesFormHandleEvent( EventPtr event );
void LoadSelTable( FormPtr pForm );
void InitSelTable( FormPtr pForm );
void SelTableDrawItem( void *table, Int16 row, Int16 column, RectanglePtr bounds );
Err WriteDBData( const void *dataP, UInt32 *sizeP, void *userDataP );


#if 0
/* Display integer data in alert while debugging. */
void DebugShowIntegerData( Coord x1, Coord y1, Coord x2, Coord y2 )
{
    Char szMessage[ 50 ];

    StrPrintF( szMessage, "(%d,%d) (%d,%d)", x1, y1, x2, y2 );
    FrmCustomAlert( UntitledAlert, szMessage, "", "" );
}
void DebugShowLongData( UInt32 lNum )
{
    Char szMessage[ 8 ];

    StrPrintF( szMessage, "%ld", lNum );
    FrmCustomAlert( UntitledAlert, szMessage, "", "" );
}
void DebugStringAndNumber( Char * szString, UInt16 nNumber )
{
    Char szMessage[ 8 ];

    StrPrintF( szMessage, "%d", nNumber );
    FrmCustomAlert( UntitledAlert, szString, szMessage, "" );
}
#endif


/* Update the global variable storing the current time. */
void UpdateCurrentTime( void )
{
    UInt32 uSecs;
    DateTimeType dateTime;


    uSecs = TimGetSeconds();
    /* Go 'back' in time to account for nMinsFromTerminus. */
    uSecs -= ( sPrefs.nMinsFromTerminus * 60 );
    TimSecondsToDateTime( uSecs, &dateTime );

    /* Generate a number of the format hhmm - between 0 and 2359. */
    gnCurrentTime = dateTime.hour * 100 + dateTime.minute;
}


Int16 GetOffset( const TablePtr pTable, const Int16 nRow, const Int16 nColumn )
{
    return ( CELL_DATA_SPACE * TblGlueGetNumberOfColumns( pTable ) * nRow ) + ( CELL_DATA_SPACE * nColumn );
}


void TableDrawItemCallback( TableType *pTable, Int16 row, Int16 column, RectangleType* bounds )
{
    TablePtr tableP = ( TablePtr ) pTable;
    UInt16 data/*, i*/;
    Char timeStr[ timeStringLength ], szFullTimeStr[ timeStringLength + CELL_DATA_SPACE ];
    Char* pCellData;
    RectangleType rect;
    UInt8 Hour, Minute;

    /* Erase previous label */
    rect = *bounds; 
    rect.topLeft.x += 1;
    rect.topLeft.y += 1;
    rect.extent.y  -= 1;
    WinEraseRectangle( &rect, 0 );

    data = TblGetItemInt( tableP, row, column );

    // TODO: Revisit this code - is 65535 condition relevant? Is 0 a valid value??
    if ( ( data == BLANK_CELL ) || ( data == 0 ) )
    {
        /* no data was found */
        StrPrintF( timeStr," " );
    }
    else if ( data == 65535 )
    {
        /* No train */
        StrPrintF( timeStr, " -- " );
    }
    else
    {
        /* Convet 713 -> 7:13. */
        Hour = data / 100;
        Minute = data % 100;
        /* Convert numbers to time format and use the user selected time format. */
        TimeToAscii( Hour, Minute, PrefGetPreference( prefTimeFormat ), timeStr );
        pCellData = MemHandleLock( pTableData );
        ErrFatalDisplayIf( pCellData == NULL, "Could not get handle to pCellData." );

        if ( *( pCellData + GetOffset( tableP, row, column ) ) != NULL )
            StrPrintF( szFullTimeStr, "%s %s", timeStr, pCellData + GetOffset( tableP, row, column ) );
        else
            StrCopy( szFullTimeStr, timeStr );
        MemHandleUnlock( pTableData );

        /* Draw new label */
        FntSetFont( stdFont );
        WinDrawChars( szFullTimeStr, StrLen( szFullTimeStr ), bounds->topLeft.x+2, bounds->topLeft.y );
    }
}


void LoadMainFormTable( FormPtr pForm )
{
    TablePtr pBusTimes;
    ListPtr pList;
    Int16 i, nColumn, numRows, nNextBusTime, nNumBuses;
    Char* pStartStr = 0, *pEndStr = 0;
    Char *pRec, *pCellData, *pPostfix;
    Char ch;
    FieldPtr pFld;
    MemHandle hRecHandle;
    RectangleType r;
    Int32 yoff;
    Boolean bMoreAvailable;    /* Maybe make a global and change to Int16 to store time of last displayed bus. */
    Coord cColWidth;
    DmOpenRef dbOpenRef;


    dbOpenRef = DmOpenDatabase( sPrefs.CardNo, sPrefs.PdbID, dmModeReadOnly );
    /* Change the table contents to reflect the selected route. */
    pBusTimes = FrmGetObjectPtr( pForm, FrmGetObjectIndex( pForm, MainFormBusTimesTable ) );
    numRows = TblGetNumberOfRows( pBusTimes );
    TblGetBounds( pBusTimes, &r );
    WinEraseRectangle( &r, 0 );

    /* Allocate memory to the cell data variable if necessary. */
    if ( pTableData == NULL )
        pTableData = MemHandleNew( CELL_DATA_SPACE * TblGetNumberOfRows( pBusTimes ) * TblGlueGetNumberOfColumns( pBusTimes ) );
    ErrFatalDisplayIf( pTableData == NULL, "Could not allocate pTableData." );
    pCellData = MemHandleLock( pTableData );
    ErrFatalDisplayIf( pCellData == NULL, "Could not get handle to pCellData." );
    MemSet( pCellData, CELL_DATA_SPACE * TblGetNumberOfRows( pBusTimes ) * TblGlueGetNumberOfColumns( pBusTimes ), 0 );
    MemHandleUnlock( pTableData );

    pList = FrmGetObjectPtr( pForm, FrmGetObjectIndex( pForm, MainFormRouteList ) );
    sPrefs.nRoute = LstGetSelection( pList );


    /* Set the To/From text fields. Use FldSetTextPtr because the it is not editable. */
    pFld = FrmGetObjectPtr( pForm, FrmGetObjectIndex( pForm, MainFormFromField ) );
    FldFreeMemory( pFld );
    if ( sPrefs.bDirection == DIRECTION_OUT )
        hRecHandle = DmQueryRecord( dbOpenRef, sPrefs.nRoute * ROUTE_REC_SIZE + FROM_OFFSET );
    else
        hRecHandle = DmQueryRecord( dbOpenRef, sPrefs.nRoute * ROUTE_REC_SIZE + TO_OFFSET );
    FldSetTextPtr( pFld, MemHandleLock( hRecHandle ) );
    FldRecalculateField( pFld, true );
    MemHandleUnlock( hRecHandle );
    FldDrawField( pFld );

    pFld = FrmGetObjectPtr( pForm, FrmGetObjectIndex( pForm, MainFormToField ) );
    FldFreeMemory( pFld );
    if ( sPrefs.bDirection == DIRECTION_OUT )
        hRecHandle = DmQueryRecord( dbOpenRef, sPrefs.nRoute * ROUTE_REC_SIZE + TO_OFFSET );
    else
        hRecHandle = DmQueryRecord( dbOpenRef, sPrefs.nRoute * ROUTE_REC_SIZE + FROM_OFFSET );
    FldSetTextPtr( pFld, MemHandleLock( hRecHandle ) );
    FldRecalculateField( pFld, true );
    MemHandleUnlock( hRecHandle );
    FldDrawField( pFld );


    /* Add times to the table. */
    pList = FrmGetObjectPtr( pForm, FrmGetObjectIndex( pForm, MainFormDayList ) );
    sPrefs.nDay = LstGetSelection( pList );
    /* Notice that sPrefs.nDay = sPrefs.nDay * 2. Each day has two directions. */
    switch ( sPrefs.nDay )
    {
        case 0: sPrefs.nDay = 0; break;   /* Mon-Fri */
        case 1: sPrefs.nDay = 2; break;   /* Saturday */
        case 2: sPrefs.nDay = 4; break;   /* Sun */
    }

    if ( sPrefs.bDirection == DIRECTION_OUT )
        hRecHandle = DmQueryRecord( dbOpenRef, sPrefs.nRoute * ROUTE_REC_SIZE + MF_OUT_TIMES_OFFSET + sPrefs.nDay );
    else
        hRecHandle = DmQueryRecord( dbOpenRef, sPrefs.nRoute * ROUTE_REC_SIZE + MF_BACK_TIMES_OFFSET + sPrefs.nDay );
    pRec = MemHandleLock( hRecHandle );


    switch ( nNextBusOffset )
    {
        /* Show buses from the start of the day. */
        case FIRST:
            pStartStr = pRec;
            pEndStr = pRec;
            /* We're already viewing the first buses so hide the 'Earlier' button. */
            FrmHideObject( pForm, FrmGetObjectIndex( pForm, MainFormEarlierButton ) );
            break;
        case NEXT:
            pStartStr = pRec;
            pEndStr = pRec;

            /* Find the next bus by looking for a bus at or after the current time. */
            nNextBusTime = 0;

            while ( *pStartStr && ( gnCurrentTime > nNextBusTime ) )
            {
                nNextBusTime = StrAToI( pStartStr );
                if ( ( gnCurrentTime > nNextBusTime ) && ( *pStartStr ) )
                {
                    /* Move to the next comma. */
                    while ( *pStartStr && *pStartStr != TIME_SEP_CHAR )
                        pStartStr++;
                    /* Move beyond the comma. */
                    if ( *pStartStr )
                        pStartStr++;
                }
            }

// TODO: Is pEndStr needed anymore?
            pEndStr = pStartStr;

            if ( pRec == pStartStr )
                /* Already viewing the first buses so hide the 'Earlier' button. */
                FrmHideObject( pForm, FrmGetObjectIndex( pForm, MainFormEarlierButton ) );
            else
                /* Not viewing first buses so show the 'Earlier' button. */
                FrmShowObject( pForm, FrmGetObjectIndex( pForm, MainFormEarlierButton ) );
            break;
        default:
            /* Begin a certain offset into the string. Used when 'Later' button pressed. */
            pStartStr = pRec + nNextBusOffset;
            pEndStr = pStartStr;
            /* Not viewing first buses so show the 'Earlier' button. */
            FrmShowObject( pForm, FrmGetObjectIndex( pForm, MainFormEarlierButton ) );
            break;
    }

    /* Count the number of bus times displayed to determine whether */
    /* a 'Next' or 'More' button is needed. */
    nNumBuses = 0;

    /* Put the times in the table. */
    i = 0;
    nColumn = 0;
    while ( *pEndStr )
    {
        while ( *pEndStr && *pEndStr != TIME_SEP_CHAR )
            pEndStr++;
        StrNCopy( szEndPoint, pStartStr, pEndStr - pStartStr );
        szEndPoint[ pEndStr - pStartStr ] = chrNull;
        if ( *pEndStr )
        {
            pEndStr++;    /* Move beyond the comma. */
            pStartStr = pEndStr;
        }

        /* Add the numeric data to the Int part of the table cell. */
        TblSetItemInt( pBusTimes, i, nColumn, StrAToI( szEndPoint ) );
        
        /* Put the data after the time, if any, into the table ptr item. */
        sPrefs.nDay = 0;
        ch = szEndPoint[ sPrefs.nDay ];

        /* Go through string looking for the first non-numeric char. */
        while ( ch && ch >= '0' && ch <= '9' )
        {
            sPrefs.nDay++;
            ch = szEndPoint[ sPrefs.nDay ];
        }
        if ( ch )
        {
            pPostfix = szEndPoint + sPrefs.nDay;
        }
        else
            pPostfix = NULL;

        pCellData = MemHandleLock( pTableData );
        ErrFatalDisplayIf( pCellData == NULL, "Could not get handle to pCellData." );
        if ( pPostfix != NULL )
            StrCopy( pCellData + GetOffset( pBusTimes, i, nColumn ), pPostfix );
        MemHandleUnlock( pTableData );    /* Added for OS5 'over locked' error. */


        nNumBuses++;

        i++;
        if ( i == numRows )
        {
            i = 0;
            nColumn++;
        }
        if ( nColumn >= TblGlueGetNumberOfColumns( pBusTimes ) )
            break;
    }
    MemHandleUnlock( hRecHandle );

    /* Table has been filled with data so there must be more buses. */
    if ( *pEndStr )
    {
        bMoreAvailable = true;
        nNextBusOffset = pEndStr - pRec;
        /* Make the 'Later' button visible. */
        FrmShowObject( pForm, FrmGetObjectIndex( pForm, MainFormLaterButton ) );
    }
    else
        /* At the end of the string therefore no more buses so hide the 'Later' button. */
        FrmHideObject( pForm, FrmGetObjectIndex( pForm, MainFormLaterButton ) );

    /* Blank out the rest of the cells by putting BLANK_CELL in them. */
    while ( nColumn < TblGlueGetNumberOfColumns( pBusTimes ) )
    {
        TblSetItemInt( pBusTimes, i, nColumn, BLANK_CELL );
        i++;
        if ( i == numRows )
        {
            i = 0;
            nColumn++;
        }
    }

    TblDrawTable( pBusTimes );

    /* Draw the horizontal lines between the rows. */
    for ( i = 1; i < numRows; i++ )
    {
        yoff = r.topLeft.y + ( i * TblGetRowHeight( pBusTimes, i ) ) - 1;
        WinDrawGrayLine( r.topLeft.x, yoff, r.extent.x, yoff );
    }

    /* Draw vertical lines between the columns. */
    for ( i = 0; i < TblGlueGetNumberOfColumns( pBusTimes ); i++ )
    {
        cColWidth = TblGetColumnWidth ( pBusTimes, i );
        WinDrawGrayLine( r.topLeft.x + cColWidth, r.topLeft.y, r.topLeft.x + cColWidth, r.topLeft.y + r.extent.y );
        r.topLeft.x += cColWidth;
    }

    /* Draw border around the table. */
    TblGetBounds( pBusTimes, &r );
    WinDrawRectangleFrame( simpleFrame, &r );

    if ( ! nNumBuses )
    {
        /* Get route number for inclusion in the alert. */
        hRecHandle = DmQueryRecord( dbOpenRef, sPrefs.nRoute * ROUTE_REC_SIZE + ROUTE_NUM_OFFSET );
        pStartStr = MemHandleLock( hRecHandle );

        /* If there are no buses and we are looking at start of timetable or times record is empty then there are no buses on that day. */
        if ( ( nNextBusOffset == FIRST ) || ( StrLen( pRec ) == 0 ) )
        {
            FrmCustomAlert( NoServiceAlert, pStartStr, LstGetSelectionText( pList, LstGetSelection( pList ) ), "" );
        }
        else
        {
            /* If simply no more buses inform the user and then show them the first buses. */
            FrmCustomAlert( NoMoreBusesAlert, pStartStr, LstGetSelectionText( pList, LstGetSelection( pList ) ), "" );
            nNextBusOffset = FIRST;
            nLastViewOffset = nNextBusOffset;
            LoadMainFormTable( pForm );
        }
        MemHandleUnlock( hRecHandle );
    }

    DmCloseDatabase( dbOpenRef );
    return;
}


/***********************************************************************
 *
 * FUNCTION:	MainFormHandleEvent
 *
 * DESCRIPTION:	Handles processing of events for the 'main' form.
 *
 * PARAMETERS:	event - the most recent event.
 *
 * RETURNED:	True if the event is handled, false otherwise.
 *
 ***********************************************************************/
Boolean MainFormHandleEvent( EventPtr event )
{
    Boolean bHandled = false;
    EventType evNewEvent;
    FormPtr frm;
    ListPtr popupList;
    UInt16 i, x;
    DmOpenRef dbOpenRef;
    TablePtr pBusTimes;
    UInt16 nNumRows;

    /* For database queries. */
    Char *pRec, *pStr;
    MemHandle hRecHandle, hStr;
    UInt16 uLen;

    Char szNumRoutes[ 5 ], szMaxRoutes[ 5 ];
    /* For Days popup list. */
    static Char Days[ 3 ][ MAX_DAY_LENGTH ];
    static Char *DayList[ 3 ];    /* Passed to LstSetListChoices. */
    UInt32 uSecs;
    DateTimeType dateTime;
    /* For Routes popup list. */
    static Char Routes[ MAX_ROUTES ][ MAX_ROUTE_NUM_LEN + 1 ];
    static Char *RouteList[ MAX_ROUTES ];    /* Passed to LstSetListChoices. */
    UInt16 nNumRoutes, nRoutesInDatabase;


    switch ( event->eType )
    {
        case ctlSelectEvent:  /* A control button was pressed and released. */
            switch ( event->data.ctlEnter.controlID )
            {
                case MainFormRouteInfoButton:
                    /* Get and display route info from the database if any is available. */
                    dbOpenRef = DmOpenDatabase( sPrefs.CardNo, sPrefs.PdbID, dmModeReadOnly );
                    if ( dbOpenRef )
                    {
                        frm = FrmGetActiveForm();
                        popupList = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormRouteList ) );
                        i = LstGetSelection( popupList );
                        popupList = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormRouteList ) );
                        hRecHandle = DmQueryRecord( dbOpenRef, i * ROUTE_REC_SIZE + ROUTE_INFO_OFFSET );
                        pRec = MemHandleLock( hRecHandle );
                    
                        if ( StrLen( pRec ) )
                        {
                            /* Copy the record to a string so that the carriage returns can be inserted. */
                            hStr = MemHandleNew( StrLen( pRec ) + 1 );
                            ErrFatalDisplayIf( NULL == hStr, "Memory allocation failed." );
                            pStr = MemHandleLock( hStr );
                            StrCopy( pStr, pRec );
                            MemHandleUnlock( hRecHandle );

                            /* Convert '|' to '\n'. */
                            uLen = StrLen( pStr );
                            for ( i = 0; i < uLen; i++ )
                                if ( *(pStr + i ) == '|' )
                                    *( pStr + i ) = '\n';
                        }
                        else
                        {
                            /* Retrieve the 'there is no info' string. */
                            hStr = DmGetResource( strRsc, NoRouteInfoString );
                            pStr = MemHandleLock( hStr );
                        }
                            
                        FrmCustomAlert( RouteInfoAlert, LstGetSelectionText( popupList, LstGetSelection( popupList ) ), pStr, "" );
                        MemHandleUnlock( hStr );

                        /* Release the resource string when no info available; free memory if there was. */
                        if ( ! StrLen( pRec ) )
                        {
                            DmReleaseResource( hStr );
                            MemHandleUnlock( hRecHandle );
                        }
                        else
                            MemHandleFree( hStr );

                        DmCloseDatabase( dbOpenRef );

                    }
                    // TODO: Handle not being able to open database.
                    //else
                    //{
                    //}

                    bHandled = true;
                    break;

                case MainFormReverseButton:
                    sPrefs.bDirection = !sPrefs.bDirection;
                    UpdateCurrentTime();
                    /* Looking at a new direction so default to viewing next buses (based on current time). */
                    nNextBusOffset = NEXT;
                    nLastViewOffset = nNextBusOffset;
                    
                    LoadMainFormTable( FrmGetActiveForm() );
                    bHandled = true;
                    break;

                /* Reset the offset so the first buses of the day are shown. */
                case MainFormEarlierButton:
                    nNextBusOffset = FIRST;
                    nLastViewOffset = nNextBusOffset;
                    
                    LoadMainFormTable( FrmGetActiveForm() );
                    bHandled = true;
                    break;

                /* Use the offset to view the next batch of buses. */
                case MainFormLaterButton:
                    nLastViewOffset = nNextBusOffset;    /* nNextBusOffset will be changed in LoadMainFormTable(). */
                    LoadMainFormTable( FrmGetActiveForm() );
                    bHandled = true;
                    break;
	   	    }
			break;

        case keyDownEvent:
            /* If a physical key */
            if ( event->data.keyDown.modifiers & commandKeyMask )
            {
                /* Send event to click Earlier button when up key pressed. */
                if ( event->data.keyDown.chr == chrPageUp )
                {
                    MemSet( &evNewEvent, sizeof( EventType ), 0 );
                    evNewEvent.eType = ctlSelectEvent;
                    evNewEvent.data.ctlSelect.controlID = MainFormEarlierButton;
                    evNewEvent.data.ctlSelect.on = true;
                    EvtAddEventToQueue( &evNewEvent );
                    bHandled = true;
                }
                /* Send event to click Later button when down key pressed. */
                else if ( event->data.keyDown.chr == chrPageDown )
                {

                    MemSet( &evNewEvent, sizeof( EventType ), 0 );
                    evNewEvent.eType = ctlSelectEvent;
                    evNewEvent.data.ctlSelect.controlID = MainFormLaterButton;
                    evNewEvent.data.ctlSelect.on = true;
                    EvtAddEventToQueue( &evNewEvent );
                    bHandled = true;
                }
            }
            break;


  		case frmOpenEvent:	/* The form was told to open. */
  		    UpdateCurrentTime();
            frm = FrmGetActiveForm();

            /* Set up Day popup list. */
            /* Mon-Fri string. */
            SysCopyStringResource( Days[ 0 ], MonFriString );
            DayList[ 0 ] = Days[ 0 ];
            /* Sat string. */
            SysCopyStringResource( Days[ 1 ], SaturdayString );
            DayList[ 1 ] = Days[ 1 ];
            /* Sun string. */
            SysCopyStringResource( Days[ 2 ], SundayString );
            DayList[ 2 ] = Days[ 2 ];
            popupList = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormDayList ) );
            LstSetListChoices( popupList, DayList, 3 );

            /* Determine day in order to set default day in list. */
            uSecs = TimGetSeconds();
            TimSecondsToDateTime( uSecs, &dateTime );

            switch ( dateTime.weekDay )
            {
                case 0: i = 2; break;    /* Sunday */
                case 6: i = 1; break;    /* Saturday */
                default: i = 0; break;   /* Mon-Fri */
            }
            LstSetSelection( popupList, i/*sPrefs.nDay*/ );
            CtlSetLabel( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormDayPopTrigger ) ), DayList[ i/*sPrefs.nDay*/ ] );


            /* Load routes into route list. */
            dbOpenRef = DmOpenDatabase( sPrefs.CardNo, sPrefs.PdbID, dmModeReadOnly );
            if ( dbOpenRef )
            {
                nNumRoutes = DmNumRecords( dbOpenRef ) / ROUTE_REC_SIZE;
                nRoutesInDatabase = nNumRoutes;

                /* Set to only read max number of routes, ignoring the rest. */
                if ( nNumRoutes >= MAX_ROUTES )
                    nNumRoutes = MAX_ROUTES - 1;

                // Use DmDatabaseSize instead of DmNumRecords because the latter generates a
                // warning in the Palm Simulator. Note that nNumRoutes must be changed to UInt32.
                //DmDatabaseSize( sPrefs.CardNo, sPrefs.PdbID, &nNumRoutes, NULL, NULL );
                //nNumRoutes = nNumRoutes / ROUTE_REC_SIZE;
                
                for ( i = 0; i < nNumRoutes; i++ )
                {
                    hRecHandle = DmQueryRecord( dbOpenRef, i * ROUTE_REC_SIZE );
                    StrCopy( Routes[ i ] , MemHandleLock( hRecHandle ) );
                    RouteList[ i ] = Routes[ i ];
                    MemHandleUnlock( hRecHandle );
                }
                DmCloseDatabase( dbOpenRef );

                /* Setup Routes popup list. */
                popupList = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormRouteList ) );
                LstSetListChoices( popupList, RouteList, nNumRoutes );

                /* Shorten the height of the list so that there are no empty 
                   spaces. The default height is 3 items. */
                if ( nNumRoutes < 3 )
                    LstSetHeight( popupList, nNumRoutes );
                else
                    LstSetHeight( popupList, 3 );

                /* Set default item in list. */
                LstSetSelection( popupList, sPrefs.nRoute );
                CtlSetLabel( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormRoutePopTrigger ) ), LstGetSelectionText( popupList, sPrefs.nRoute ) );

            }
            // TODO: Quit app if problem
            //else
            //{
            //}

            pBusTimes = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormBusTimesTable ) );
            nNumRows = TblGetNumberOfRows( pBusTimes );

            for ( i = 0; i < nNumRows; i++ )
            {
                TblSetRowSelectable( pBusTimes, i, false );    /* Don't allow row to be selected. */
                /* Set row types - custom type. Painting will be handled by a callback function. */
                for ( x = 0; x < TblGlueGetNumberOfColumns( pBusTimes ); x++ )
                    TblSetItemStyle( pBusTimes, i, x, customTableItem );
            }

            /* Set our custom table item draw callback & ensure columns are usable (aka visible). */
            for ( i = 0; i < TblGlueGetNumberOfColumns( pBusTimes ); i++ )
            {
                TblSetColumnUsable( pBusTimes, i, true );
                TblSetCustomDrawProcedure( pBusTimes, i, (TableDrawItemFuncPtr)TableDrawItemCallback );
            }

            /* Form has already been loaded and activated so just draw it. */
            FrmDrawForm( frm );

            /* Load data into the table. */
            //nNextBusOffset = NEXT;    /* Display starting at current time when form first opened. */
            nNextBusOffset = nLastViewOffset;
            LoadMainFormTable( frm );

            /* Display a message if db has more routes than app can handle. */
            if ( nRoutesInDatabase >= MAX_ROUTES )
            {
                StrPrintF( szNumRoutes, "%d", nRoutesInDatabase );
                StrPrintF( szMaxRoutes, "%d", MAX_ROUTES );
                FrmCustomAlert( DatabaseTooManyRoutesAlert, szNumRoutes, szMaxRoutes, "" );
            }

            bHandled = true;
            break;

	    case frmCloseEvent:
                // Uncommenting this line in seems to cause a memory leak somehow.
                //bHandled = true;

	        break;

		case menuEvent:		/* A menu item was selected. */
                    /* First clear the menu status from the display. */
                    MenuEraseStatus( 0 );

                    switch ( event->data.menu.itemID )
                    {
                        case MainFormAboutBusSched:
                            /* Load the About form, then display it. */
                            frm = FrmInitForm( AboutForm );
                            FrmDoDialog( frm );

                            /* Delete the About form. */
                            FrmDeleteForm( frm );

                            bHandled = true;
                            break;

                        case MainFormSelecttimetable:
                            // Was causing 'Form already loaded' warning in Simulator.
                            // Probably already loaded by ApplicationHandleEvent().
                            //frm = FrmInitForm( SelectTimetableForm );
                            FrmGotoForm( SelectTimetableForm );

                            bHandled = true;
                            break;

                        case MainFormPreferences:
                            FrmGotoForm( PreferencesForm );

                            bHandled = true;
                            break;
                    }

                    if ( bHandled )
                        break;

	    /* Received when this window/form becomes the active window. */
/*	    case winEnterEvent:
//            bHandled = true;
            break;
*/
            case popSelectEvent:
                switch( event->data.popSelect.listID )
                {
                    case MainFormRouteList:
                        UpdateCurrentTime();
                        /* Looking at a new route so default to viewing next buses (based on current time). */
                        nNextBusOffset = NEXT;
                        nLastViewOffset = nNextBusOffset;

                        frm = FrmGetActiveForm();

                        /* Hack to get trigger/list to display selected item. */
                        /* This ensures the route list is updated before any messages */
                        /* with the new route number in them are displayed. */
                        /* Also so UI shows correct route for any error messages */
                        popupList = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormRouteList ) );
                        LstSetSelection( popupList, event->data.popSelect.selection );
                        CtlSetLabel( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormRoutePopTrigger ) ), LstGetSelectionText( popupList, event->data.popSelect.selection ) );

                        LoadMainFormTable( frm );

                        bHandled = true;
                        break;

                    case MainFormDayList:
                        UpdateCurrentTime();
                        nNextBusOffset = NEXT;    /* Looking at a new route so default to viewing next buses (based on current time). */
                        nLastViewOffset = nNextBusOffset;

                        frm = FrmGetActiveForm();

                        /* Hack to get trigger/list to display selected item. */
                        /* This ensures the day list is updated before any */
                        /* messages with the new day in them are displayed. */
                        /* Also so UI shows correct day for any error messages */
                        popupList = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormDayList ) );
                        LstSetSelection( popupList, event->data.popSelect.selection );
                        CtlSetLabel( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, MainFormDayPopTrigger ) ), LstGetSelectionText( popupList, event->data.popSelect.selection ) );

                        LoadMainFormTable( frm );

                        bHandled = true;
                        break;
                    }

                default:
	            break;
        }

        return bHandled;
}


/* This function allows the selection of a new timetable database. */
Boolean SelectTimetableFormHandleEvent( EventPtr event )
{
    Boolean bHandled = false;
    FormPtr frm;


    switch (event->eType)
    {
        case ctlSelectEvent:  /* A control button was pressed and released. */
            switch ( event->data.ctlEnter.controlID )
            {
                case SelectTimetableOKButton:
                    {
                    TablePtr pTable;
                    Int16 nRow, nColumn;
                    dbRowType *drtData = NULL;
                    MemHandle hText;


                    frm = FrmGetActiveForm();
                    pTable = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, SelectTimetableSelectTimetableTable ) );

                    if ( true == TblGetSelection( pTable, &nRow, &nColumn ) )
                    {
                        drtData = ( dbRowType * ) TblGetItemPtr( pTable, nRow, nColumn );
                        if ( drtData )
                        {
                            sPrefs.CardNo = drtData->CardNo;
                            sPrefs.PdbID = drtData->PdbID;
                            nNextBusOffset = NEXT;    /* Default to displaying next bus in new timetable. */
                            nLastViewOffset = nNextBusOffset;
                        }

                    }
                    else
                    {
                        /* You must select a database before clicking the 'OK' button. */
                        hText = DmGetResource( strRsc, NoSelectionString );
                        FrmCustomAlert( ErrorAlert, ( Char * ) MemHandleLock( hText ), "", "" );
                        MemHandleUnlock( hText );
                        DmReleaseResource( hText );

                        /* This flags code below to stay in this form. */
                        bHandled = true;
                    }
                    }

                    /* Stay in this form if nothing was selected. */
                    if ( ! bHandled )
                        FrmGotoForm( MainFormForm );
                    bHandled = true;
                    break;

                case SelectTimetableCancelButton:
                    FrmGotoForm( MainFormForm );
                    bHandled = true;
                    break;

                case SelectTimetableBeamButton:
                    {
                    TablePtr pTable;
                    Int16 nRow, nColumn;
                    dbRowType *drtData = NULL;
                    MemHandle hText;
                    ExgSocketType exgSocket;
                    Err err = 0;


                    frm = FrmGetActiveForm();
                    pTable = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, SelectTimetableSelectTimetableTable ) );

                    if ( true == TblGetSelection( pTable, &nRow, &nColumn ) )
                    {
                        drtData = ( dbRowType * ) TblGetItemPtr( pTable, nRow, nColumn );
                        if ( drtData )
                        {
// TODO: Figure out how this code works.
                            MemSet(&exgSocket, sizeof(exgSocket),0);
                            exgSocket.description = drtData->DBName;
                            exgSocket.name = "BusSched.pdb";

                            err = ExgPut( &exgSocket );
                            if ( !err )
                            {
                                err = ExgDBWrite( WriteDBData, &exgSocket, drtData->DBName, drtData->PdbID, drtData->CardNo );
                                err = ExgDisconnect( &exgSocket, err );
                            }
                        }
                    }
                    else
                    {
                        /* You must select a database to beam. */
                        hText = DmGetResource( strRsc, NoSelectionString );
                        FrmCustomAlert( ErrorAlert, ( Char * ) MemHandleLock( hText ), "", "" );
                        MemHandleUnlock( hText );
                        DmReleaseResource( hText );
                    }
                    }

                    bHandled = true;
                    break;

                case SelectTimetableInfoButton:
                    {
                    TablePtr pTable;
                    Int16 nRow, nColumn;
                    dbRowType *drtData = NULL;
                    MemHandle hText;
                    DmOpenRef dbOpenRef;
                    MemHandle hInfoRecord;
                    Char *pInfoRecord;


                    frm = FrmGetActiveForm();
                    pTable = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, SelectTimetableSelectTimetableTable ) );

                    if ( true == TblGetSelection( pTable, &nRow, &nColumn ) )
                    {
                        drtData = ( dbRowType * ) TblGetItemPtr( pTable, nRow, nColumn );
                        if ( drtData )
                        {
                            hText = 0;
                            hInfoRecord = 0;
                            /* Use the text of the last record (short info text), or the db name on error. */
                            pInfoRecord = '\0';    /* If pInfoRecord is not set it will look like zero length string. */
                            dbOpenRef = DmOpenDatabase( drtData->CardNo, drtData->PdbID, dmModeReadOnly );
                            if ( dbOpenRef )
                            {
                                /* Getting record right after the last route block. */
                                hInfoRecord = DmQueryRecord( dbOpenRef, ( DmNumRecords( dbOpenRef ) / ROUTE_REC_SIZE ) * ROUTE_REC_SIZE );

                                /* Handle when the record doesn't exist or is empty. */
                                if ( hInfoRecord != NULL )
                                {
                                    // TODO: Consider processing string to convert '|' to '\n'.
                                    FrmCustomAlert( DatabaseInfoAlert, (Char *)MemHandleLock( hInfoRecord ), "", "" );
                                    MemHandleUnlock( hInfoRecord );
                                }
                                else
                                {
                                    hText = DmGetResource( strRsc, NoDbInfoString );
                                    FrmCustomAlert( DatabaseInfoAlert, (Char *)MemHandleLock( hText ), "", "" );
                                    MemHandleUnlock( hText );
                                    DmReleaseResource( hText );
                                }
                                DmCloseDatabase( dbOpenRef );
                            }
                            else
                            {
                                hText = DmGetResource( strRsc, NoDbInfoString );
                                FrmCustomAlert( DatabaseInfoAlert, (Char *)MemHandleLock( hText ), "", "" );
                                MemHandleUnlock( hText );
                                DmReleaseResource( hText );
                            }
                        }

                    }
                    else
                    {
                        /* You must select a database to get info on. */
                        hText = DmGetResource( strRsc, NoSelectionString );
                        FrmCustomAlert( ErrorAlert, ( Char * ) MemHandleLock( hText ), "", "" );
                        MemHandleUnlock( hText );
                        DmReleaseResource( hText );
                    }
                    }

                    bHandled = true;
                    break;

                case SelectTimetableDeleteButton:
                    {
                    TablePtr pTable;
                    Int16 nRow, nColumn;
                    dbRowType *drtData = NULL;
                    MemHandle hText;


                    frm = FrmGetActiveForm();
                    pTable = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, SelectTimetableSelectTimetableTable ) );

                    if ( true == TblGetSelection( pTable, &nRow, &nColumn ) )
                    {
                        drtData = ( dbRowType * ) TblGetItemPtr( pTable, nRow, nColumn );
                        if ( drtData )
                        {
                            /* Disallow deletion of current database. */
                            if ( ( sPrefs.CardNo == drtData->CardNo ) && ( sPrefs.PdbID == drtData->PdbID ) )
                            {
                                hText = DmGetResource( strRsc, NoDelCurrentDB );
                                FrmCustomAlert( ErrorAlert, (Char *)MemHandleLock( hText ), "", "" );
                                MemHandleUnlock( hText );
                                DmReleaseResource( hText );
                            }
                            else
                            {
// TODO: Handle deletion errors.
                                DmDeleteDatabase( drtData->CardNo, drtData->PdbID );

                                /* Delete row from table and redraw the table */
                                TblRemoveRow( pTable, nRow );
                                TblUnhighlightSelection( pTable );
                                TblDrawTable( pTable );
                            }
                        }

                    }
                    else
                    {
                        /* You must select a database to delete. */
                        hText = DmGetResource( strRsc, NoSelectionString );
                        FrmCustomAlert( ErrorAlert, ( Char * ) MemHandleLock( hText ), "", "" );
                        MemHandleUnlock( hText );
                        DmReleaseResource( hText );
                    }
                    }

                    bHandled = true;
                    break;
            }
            break;

/*        case frmLoadEvent:
FrmCustomAlert( UntitledAlert, "frmLoadEvent.", "", "" );
            frm = FrmGetFormPtr( event->data.frmLoad.formID );
            FrmSetActiveForm( frm );
            bHandled = true;
            break;
*/
        case frmOpenEvent:
            frm = FrmGetActiveForm();

            /* Ensure the PdbArrayH pointer is not allocated. */
            if ( PdbArrayH != NULL )
            {
                MemHandleFree( PdbArrayH );
                PdbArrayH = NULL;
            }

            FrmDrawForm( frm );
            InitSelTable( frm );
            LoadSelTable( frm );

            bHandled = true;
            break;

        case frmCloseEvent:
//        case winExitEvent:
            frm = FrmGetActiveForm();
            
            if ( PdbArrayH != NULL )
            {
                MemHandleFree( PdbArrayH );
                PdbArrayH = NULL;
            }

            FrmEraseForm( frm );
            FrmDeleteForm( frm );

            bHandled = true;
            break;

        default:
            break;
    }

    return bHandled;
}


/* Initialize the SelectTimetableForm Table. */
void InitSelTable( FormPtr pForm )
{
    RectangleType r;
    TablePtr tableP;
    Int32 i;
    UInt16 numRows = 0;


    tableP = FrmGetObjectPtr( pForm, FrmGetObjectIndex( pForm, SelectTimetableSelectTimetableTable ) );
    numRows = TblGetNumberOfRows( tableP );

    /* Setup the rows of the table */
    for( i = 0; i < numRows; i++ )
    {
        TblSetItemStyle( tableP, i, 0, customTableItem );
        TblSetRowSelectable( tableP, i, true );
    }

    /* Make sure the columns are usable - so they draw */
    TblSetColumnUsable( tableP, 0, true );
    TblHasScrollBar( tableP, true );

    /* Set our custom table item draw callback */
    TblSetCustomDrawProcedure( tableP, 0, SelTableDrawItem );

    /* Draw the outline of the table */
    TblGetBounds( tableP, &r );

    WinDrawRectangleFrame( simpleFrame, &r );
}


/* Populates and draw selectForm table */
void LoadSelTable( FormPtr pForm )
{
    TablePtr tableP;
    Int16 i = 0;
    Int16 PdbCount = 0;
    UInt16 numRows = 0;
    Boolean newSearch = true;
    DmSearchStateType searchState;
    UInt16 CardNumber;
    LocalID dbID;
    Char name[ dmDBNameLength ];
    dbRowType *PdbArrayP;
#ifdef DISPLAY_DB_INFO_IN_SEL_FORM
    DmOpenRef dbOpenRef;
    MemHandle hInfoRecord;
    Char *pInfoRecord;
#endif


    tableP = FrmGetObjectPtr( pForm, FrmGetObjectIndex( pForm, SelectTimetableSelectTimetableTable ) );
    numRows = TblGetNumberOfRows( tableP );

    /* Let's count how many databases we have in memory. */
    while ( 0 == DmGetNextDatabaseByTypeCreator( newSearch, &searchState, PDBTYPE, CREATORID, false, &CardNumber, &dbID ) )
    {
        PdbCount++;
        newSearch = false;
    }

    if ( PdbCount > 0 )
    {
        PdbArrayH = MemHandleNew( PdbCount * sizeof( dbRowType ) );
        PdbArrayP = MemHandleLock( PdbArrayH );

        newSearch = true;
        PdbCount = 0;

        while ( 0 == DmGetNextDatabaseByTypeCreator( newSearch, &searchState, PDBTYPE, CREATORID, false, &CardNumber, &dbID ) )
        {
            newSearch = false;

            DmDatabaseInfo( CardNumber, dbID, name, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );

#ifdef DISPLAY_DB_INFO_IN_SEL_FORM
            /* Use the text of the last record (short info text), or the db name on error. */
            dbOpenRef = DmOpenDatabase( CardNumber, dbID, dmModeReadOnly );
            if ( dbOpenRef )
            {
                hInfoRecord = DmQueryRecord( dbOpenRef, DmNumRecords( dbOpenRef ) - 1 );
                pInfoRecord = MemHandleLock( hInfoRecord );
                MemMove( PdbArrayP[ PdbCount ].DBName, pInfoRecord, dmDBNameLength/*StrLen( pInfoRecord )*/ );
                MemHandleUnlock( hInfoRecord );
                DmCloseDatabase( dbOpenRef );
            }
            else
                MemMove( PdbArrayP[ PdbCount ].DBName, name, dmDBNameLength );
#else
            /* Simply list the datbase name in the table. */
            MemMove( PdbArrayP[ PdbCount ].DBName, name, dmDBNameLength );
#endif

            /* Append ' (*)' to indicate selected database. */
            // TODO: Consider putting this string into resources.
            // TODO: Ensure this string doesn't make DBName exceed dmDBNameLength.
            if ( ( sPrefs.CardNo == CardNumber ) && ( sPrefs.PdbID == dbID ) )
                MemMove( PdbArrayP[ PdbCount ].DBName + StrLen(name), " (*)\0", 5 );

            PdbArrayP[ PdbCount ].CardNo = CardNumber;
            PdbArrayP[ PdbCount ].PdbID = dbID;

            PdbCount++;
        }

        while ( ( i < numRows ) && ( i < PdbCount ) )
        {
            TblSetItemPtr( tableP, i, 0, ( dbRowType* ) PdbArrayP++ );
            i++; 
        }

// DCARBERY: I don't have a scroll bar yet.
//    SclSetScrollBar(FrmGetObjectPtr( pForm, 
//                     FrmGetObjectIndex( pForm, 
//                      selectFrmTableScrollBar )), 
//                      0,          /* current value */
//                      0,          /* min */
//                      8,          /* max */
//                      numRows );  /* page size */

        TblDrawTable( tableP );
    }
}


/* Draws items in selection table */
void SelTableDrawItem( void *table, Int16 row, Int16 column, RectanglePtr bounds )
{
    TablePtr tableP = ( TablePtr ) table;
    Char str[ dmDBNameLength ] = { 0 };
    RectangleType rect;
    dbRowType *data;
  
    /* Erase previous label */
    rect = *bounds; 
    rect.topLeft.x += 1;
    rect.topLeft.y += 1;
    rect.extent.y  -= 1;
    WinEraseRectangle( &rect, 0 );
  
// This is the old style, no longer supported.
//    data = (dbRowType*)tableP->items[row * tableP->numColumns + column].ptr;
    data = ( dbRowType* ) TblGetItemPtr( tableP, row, column );
  
    if ( data )
        StrCopy( str, data->DBName );
    else
        TblSetRowSelectable( tableP, row, false );

    /* Draw new label */
    FntSetFont( stdFont );
    WinDrawChars( str, StrLen( str ), bounds->topLeft.x+2, bounds->topLeft.y );
  
  // DCARBERY: Select row containing current open database.
/*  if ( StrCompare( data->DBName, DBName ) == 0 )
      TblGlueSetSelection( tableP, row, column );*/
}


/* Callback to ExgDBWrite for beaming */
Err WriteDBData( const void *dataP, UInt32 *sizeP, void *userDataP )
{
    Err err;
  
    *sizeP = ExgSend( ( ExgSocketPtr ) userDataP, ( void* )dataP, *sizeP, &err );

    return err;
}


/* Set preferences. */
Boolean PreferencesFormHandleEvent( EventPtr event )
{
    Boolean bHandled = false;
    FormPtr frm;


    switch ( event->eType )
    {
        case ctlSelectEvent:  /* A control button was pressed and released. */
            switch ( event->data.ctlEnter.controlID )
            {
                case PreferencesOKButton:
                    {
                    FieldType *pTime;
                    MemHandle hText;

                    frm = FrmGetActiveForm();
                    pTime = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, PreferencesMinutesField ) );
                    hText = FldGetTextHandle( pTime );
                    sPrefs.nMinsFromTerminus = StrAToI( MemHandleLock( hText ) );
                    MemHandleUnlock( hText );

                    FrmGotoForm( MainFormForm );
                    bHandled = true;
                    break;
                    }

                case PreferencesCancelButton:
                    FrmGotoForm( MainFormForm );
                    bHandled = true;
                    break;

            }
            break;

/*
        case frmLoadEvent:
FrmCustomAlert( UntitledAlert, "frmLoadEvent.", "", "" );
            frm = FrmGetFormPtr( event->data.frmLoad.formID );
            FrmSetActiveForm( frm );
            bHandled = true;
            break;
*/

        case frmOpenEvent:
            frm = FrmGetActiveForm();

            FrmDrawForm( frm );
            {
            FieldType *pTime;
            Char szMins[ 5 ];
            Char *pMins;
            MemHandle hText;
            
            StrPrintF( szMins, "%d", sPrefs.nMinsFromTerminus );
            hText = MemHandleNew( StrLen( szMins ) + 1 );
            pMins = MemHandleLock( hText );
            StrCopy( pMins, szMins );
            MemHandleUnlock( hText );

            pTime = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, PreferencesMinutesField ) );
            FldSetTextHandle( pTime, hText );
            FldDrawField( pTime );
            /* Set focus to the text field. */
            FrmSetFocus( frm, FrmGetObjectIndex( frm, PreferencesMinutesField ) );
            }

	   	    bHandled = true;
	   	    break;

        case frmCloseEvent:
            frm = FrmGetActiveForm();
            
            FrmEraseForm( frm );
            FrmDeleteForm( frm );

            bHandled = true;
            break;

        default:
            break;
    }

    return bHandled;
}


/***********************************************************************
 *
 * FUNCTION:    ApplicationHandleEvent
 *
 * DESCRIPTION: This routine loads a form resource and sets the event handler for the form.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    True if the event has been handled and should not be
 *				passed to a higher level handler.
 *
 ***********************************************************************/
Boolean ApplicationHandleEvent( const EventPtr event )
{
	FormPtr	frm;
	UInt16 formId;
	Boolean	handled = false;

	if ( event->eType == frmLoadEvent )
	{
		/* Load the form resource specified in the event then activate the form. */
		formId = event->data.frmLoad.formID;
		frm = FrmInitForm( formId );
		FrmSetActiveForm( frm );

		/* Set the event handler for the form.  The handler of the currently         */
		/* active form is called by FrmDispatchEvent each time it receives an event. */
		switch ( formId )
		{
			case MainFormForm:
				FrmSetEventHandler( frm, MainFormHandleEvent );
				break;

            case SelectTimetableForm:
                FrmSetEventHandler( frm, SelectTimetableFormHandleEvent );
                break;

            case PreferencesForm:
                FrmSetEventHandler( frm, PreferencesFormHandleEvent );
                break;
		}
		handled = true;
	}
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:	EventLoop
 *
 * DESCRIPTION:	A simple loop that obtains events from the Event
 *				Manager and passes them on to various applications and
 *				system event handlers before passing them on to
 *				FrmDispatchEvent for default processing.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/
void EventLoop(void)
{
    EventType event;
    UInt16 error;
	
    do
    {
        /* Get the next available event. */
        EvtGetEvent( &event, evtWaitForever );
		
        /* Give the system a chance to handle the event. */
        if (! SysHandleEvent( &event ) )
            /* Give the menu bar a chance to update and handle the event. */
            if (! MenuHandleEvent( 0, &event, &error ) )
                /* Give the application a chance to handle the event. */
                if ( ! ApplicationHandleEvent( &event ) )
                    /* Let the form object provide default handling of the event. */
                    FrmDispatchEvent( &event );
    } while ( event.eType != appStopEvent );
}


/***********************************************************************
 *
 * FUNCTION:     StartApplication
 *
 * DESCRIPTION:  This routine sets up the initial state of the application.
 *
 * PARAMETERS:   None.
 *
 * RETURNED:     Nothing.
 *
 ***********************************************************************/
void StartApplication(void)
{
    EventType newEvent;
    DmSearchStateType searchState;
    MemHandle hRecHandle;
    Boolean bNewSearch = true, bPDBExists = false;
    UInt16 nPrefsSize, nPrefsVersion;
//    sPrefsStruct sStoredPrefs;
    LocalID newID;
    UInt16 newCardNo;


    /* Get previous user selection in the preferences. */
    nPrefsSize = sizeof( sPrefsStruct );
    nPrefsVersion = PrefGetAppPreferences( CREATORID, 0, &sPrefs, &nPrefsSize, true );

    if ( nPrefsVersion != noPreferenceFound )
    {

        /* First let's verify the PDB is still available */
        while ( 0 == DmGetNextDatabaseByTypeCreator( bNewSearch, &searchState, PDBTYPE, CREATORID, false, &newCardNo, &newID ) )
        {
            bNewSearch = false;
            if ( sPrefs.PdbID == 0 )
                break;

            /* Finish search if database found. */
            if ( newID == sPrefs.PdbID )
            {
                bPDBExists = true;
                //sPrefs.PdbID = newID;
                break;
            }
        }
    }


// TODO: Get database info from Preferences.
//    sPrefs.bDirection = DIRECTION_OUT;    /* Default bus direction. */
//    sPrefs.nMinsFromTerminus = 0;    /* Default to 0 mins from terminus. */

    /* If no stored PDB or stored one doesn't exist, search for one. */
    if ( ! bPDBExists )
    {
        /* Open first database found else display an error and quit. */
        if ( 0 == DmGetNextDatabaseByTypeCreator( true, &searchState, PDBTYPE, CREATORID, false, &sPrefs.CardNo, &sPrefs.PdbID ) )
// TODO: Is it really necessary to record the datbase name? Is this code needed?
            DmDatabaseInfo( sPrefs.CardNo, sPrefs.PdbID, /*DBName*/NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
        else
        {
            /* Error message and quit app if database cannot be opened. */
            hRecHandle = DmGetResource( strRsc, NoDbsString );
            FrmCustomAlert( ErrorAlert, ( Char * ) MemHandleLock( hRecHandle ), "", "" );
            MemHandleUnlock( hRecHandle );
            DmReleaseResource( hRecHandle );

            MemSet( &newEvent, sizeof( EventType ), 0 );
            newEvent.eType = appStopEvent;
            EvtAddEventToQueue( &newEvent );
        }
    }
}


/***********************************************************************
 *
 * FUNCTION:    StopApplication
 *
 * DESCRIPTION: This routine tidies up, freeing memory and similar activies.
 *
 * PARAMETERS:  Nothing
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/
void StopApplication(void)
{
    /* Free memory block for table cell data. */
    if ( pTableData != NULL )
        MemHandleFree( pTableData );

    /* Store the user preferences. Use pref db that is backed up with Hotsync. */
    PrefSetAppPreferences( CREATORID, 0, PREFAPPVERSION, &sPrefs, sizeof( sPrefsStruct ), true );

    /* Close all open forms to allow their frmCloseEvent handlers */
    /* to execute. An appStopEvent doesn't send frmCloseEvents.   */
    FrmCloseAllForms();
}


/***********************************************************************
 *
 * FUNCTION:    RomVersionCompatible
 *
 * DESCRIPTION: Check that the ROM version meets your minumum
 *              requirement.  Warn if the app was switched to.
 *
 * PARAMETERS:  requiredVersion - minimum rom version required
 *                                (see sysFtrNumROMVersion in SystemMgr.h 
 *                                for format)
 *              launchFlags     - flags indicating how the application was
 *								  launched.  A warning is displayed only if
 *                                these flags indicate that the app is 
 *								  launched normally.
 *
 * RETURNED:    Zero if rom is compatible else an error code.
 *                             
 ***********************************************************************/
Err RomVersionCompatible( const UInt32 requiredVersion, const UInt16 launchFlags )
{
    UInt32 romVersion;
	
	
    /* See if we're on in minimum required version of the ROM or later.       */
    /* The system records the version number in a feature.  A feature is a    */
    /* piece of information which can be looked up by a creator and feature   */
    /* number.                                                                */
    FtrGet( sysFtrCreator, sysFtrNumROMVersion, &romVersion );
    if ( romVersion < requiredVersion )
    {
        /* If the user launched the app from the launcher, explain why   */
        /* the app shouldn't run. If the app was contacted for something */
        /* else, like it was asked to find a string by the system find, then  */
        /* don't bother the user with a warning dialog.  These flags tell how */
        /* the app was launched to decided if a warning should be displayed.  */
        if ( ( launchFlags & ( sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp ) ) ==
            ( sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp ) )
        {
            FrmAlert( RomIncompatibleAlert );
		
            /* Pilot 1.0 will continuously relaunch this app unless we switch to */
            /* another safe one. The sysFileCDefaultApp is considered "safe".    */
            if ( romVersion < 0x02000000 )
                AppLaunchWithCommand( sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL );
        }
		
        return ( sysErrRomIncompatible );
    }

    return 0;
}


/***********************************************************************
 *
 * FUNCTION:		PilotMain
 *
 ***********************************************************************/
UInt32 PilotMain( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags )
{
    UInt16 error;    /* Error starting the app. */

	
    /* This app makes use of PalmOS 2.0 features. It will crash if run on an   */
    /* earlier version of PalmOS.  Detect and warn if this happens, then exit. */
    /* 0x02000000 is PalmOS 2.0 version number. */
    error = RomVersionCompatible( 0x02000000, launchFlags );
    if ( error )
        return error;
	
    /* This app doesn't support any action codes. */
    if ( cmd != sysAppLaunchCmdNormalLaunch ) return sysErrParamErr;

    /* Initialize the application's data structures, ... */
    StartApplication();

    /* Start the first form */
    FrmGotoForm( MainFormForm );
		
    /* Start the event loop. */
    EventLoop();

    /* Clean up before exiting the application. */
    StopApplication();

    return 0;
}
